/*
 * This file is part of Skyward Hub.
 *
 * Skyward Hub is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Skyward Hub is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Skyward Hub. If not, see <https://www.gnu.org/licenses/>.
 *
 */

#include "CommandPad.h"

#include <Modules/CommandPad/MessagesList.h>
#include <Modules/Mavlink/MavlinkCodec.h>

#include <QVBoxLayout>

CommandPad::CommandPad() : Module(ModuleId::COMMAND_PAD) { setupUi(); }

CommandPad::~CommandPad()
{
    for (auto key : formElements.keys())
        delete formElements[key];
}

XmlObject CommandPad::toXmlObject()
{
    XmlObject obj = Module::toXmlObject();

    obj.addAttribute("target_system", systemsComboBox->currentText());

    auto key = messagesListComboBox->currentText();
    if (formElements.contains(key))
    {
        obj.addAttribute("message_id", key);
        formElements[key]->toXmlObject(obj);
    }

    return obj;
}

void CommandPad::fromXmlObject(const XmlObject &obj)
{
    auto key = obj.getAttribute("message_id");

    auto targetSystem = obj.getAttribute("target_system");
    auto index        = systemsComboBox->findText(targetSystem);
    if (index >= 0)
    {
        systemsComboBox->setCurrentIndex(index);
    }

    if (formElements.contains(key))
    {
        formElements[key]->fromXmlObject(obj);
        messagesListComboBox->setCurrentText(key);
    }
}

void CommandPad::setupUi()
{
    QVBoxLayout *outerLayout = new QVBoxLayout;
    outerLayout->addStretch();

    systemsComboBox = new QComboBox;
    systemsComboBox->setSizePolicy(
        QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed));
    systemsComboBox->addItem("Main", MavlinkCodec::SysIdMain);
    systemsComboBox->addItem("Payload", MavlinkCodec::SysIdPayload);
    systemsComboBox->addItem("Rig", MavlinkCodec::SysIdRig);
    systemsComboBox->addItem("Gs Receiver", MavlinkCodec::SysIdGsReceiver);
    outerLayout->addWidget(systemsComboBox);

    messagesListComboBox = new QComboBox;
    messagesListComboBox->setSizePolicy(
        QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed));
    messagesListComboBox->addItems(MessagesList::messagesList);
    outerLayout->addWidget(messagesListComboBox);

    QGridLayout *formGridLayout = new QGridLayout;
    outerLayout->addLayout(formGridLayout);

    QPushButton *sendButton = new QPushButton("Send");
    outerLayout->addWidget(sendButton);

    setLayout(outerLayout);

    MessagesList::fillMessagesMap(formElements);

    // Set default value
    currentMessage = "PING_TC";
    messagesListComboBox->setCurrentText(currentMessage);

    connect(messagesListComboBox, &QComboBox::currentTextChanged, this,
            [=](QString key)
            {
                if (formElements.contains(key))
                {
                    formElements[currentMessage]->removeFromGridLayout(
                        formGridLayout);
                    formElements[key]->applyToGridLayout(formGridLayout);
                    currentMessage = key;
                }
            });

    connect(sendButton, &QPushButton::clicked,
            [=]()
            {
                auto key = messagesListComboBox->currentText();
                if (formElements.contains(key))
                {
                    auto message = formElements[key]->prepareMessage(
                        systemsComboBox->currentText(), key);
                    MessageBroker::getInstance().publish(message);
                }
            });
}
